@@ -10,8 +10,10 @@
 
 #include <KEncodingProber>
 #include <KSelectAction>
-#include <memory>
 #include <kconfigwidgets_export.h>
+#include <memory>
+
+class QTextCodec;
 
 /**
  *  @class KCodecAction kcodecaction.h KCodecAction
@@ -20,8 +22,7 @@
  *
  *  This action shows up a submenu with a list of the available codecs on the system.
  */
-class KCONFIGWIDGETS_EXPORT KCodecAction
-    : public KSelectAction
+class KCONFIGWIDGETS_EXPORT KCodecAction : public KSelectAction
 {
     Q_OBJECT
 
@@ -72,7 +73,7 @@ Q_SIGNALS:
      * @deprecated Since 5.78, use codecTriggered(QTextCodec *)
      */
     KCONFIGWIDGETS_DEPRECATED_VERSION(5, 78, "Use KCodecAction::codecTriggered(QTextCodec *)")
-    void triggered(QTextCodec *codec);
+    void triggered(QTextCodec *codec); // clazy:exclude=overloaded-signal
 #endif
     /**
      * Specific (proper) codec was selected
@@ -108,7 +109,7 @@ Q_SIGNALS:
      * @deprecated Since 5.78, use encodingProberTriggered(KEncodingProber::ProberType)
      */
     KCONFIGWIDGETS_DEPRECATED_VERSION(5, 78, "Use KCodecAction::encodingProberTriggered(KEncodingProber::ProberType)")
-    void triggered(KEncodingProber::ProberType);
+    void triggered(KEncodingProber::ProberType); // clazy:exclude=overloaded-signal
 #endif
 
     /**
@@ -153,8 +154,6 @@ protected:
 private:
     friend class KCodecActionPrivate;
     std::unique_ptr<class KCodecActionPrivate> const d;
-
-    Q_PRIVATE_SLOT(d, void _k_subActionTriggered(QAction *))
 };
 
 #endif
