@@ -28,6 +28,8 @@
 #include <KConfigGroup>
 #include <KLocalizedString>
 
+#include <set>
+
 KRecentFilesAction::KRecentFilesAction(QObject *parent)
     : KSelectAction(parent)
     , d_ptr(new KRecentFilesActionPrivate(this))
@@ -75,15 +77,22 @@ void KRecentFilesActionPrivate::init()
     clearAction->setObjectName(QStringLiteral("clear_action"));
     clearAction->setVisible(false);
     q->setEnabled(false);
-    q->connect(q, SIGNAL(triggered(QAction *)), SLOT(_k_urlSelected(QAction *)));
+    q->connect(q, qOverload<QAction *>(&KSelectAction::triggered), q, [this](QAction *action) {
+        urlSelected(action);
+    });
 }
 
 KRecentFilesAction::~KRecentFilesAction() = default;
 
-void KRecentFilesActionPrivate::_k_urlSelected(QAction *action)
+void KRecentFilesActionPrivate::urlSelected(QAction *action)
 {
     Q_Q(KRecentFilesAction);
-    Q_EMIT q->urlSelected(m_urls[action]);
+
+    auto it = findByAction(action);
+
+    Q_ASSERT(it != m_recentActions.cend()); // Should never happen
+
+    Q_EMIT q->urlSelected(it->url);
 }
 
 int KRecentFilesAction::maxItems() const
@@ -98,9 +107,20 @@ void KRecentFilesAction::setMaxItems(int maxItems)
     // set new maxItems
     d->m_maxItems = std::max(maxItems, 0);
 
-    // remove all excess items
-    while (selectableActionGroup()->actions().count() > d->m_maxItems) {
-        delete removeAction(selectableActionGroup()->actions().last());
+    // Remove all excess items, oldest (i.e. first added) first
+    const int difference = static_cast<int>(d->m_recentActions.size()) - d->m_maxItems;
+    if (difference > 0) {
+        using Info = KRecentFilesActionPrivate::RecentActionInfo;
+
+        auto beginIt = d->m_recentActions.begin();
+        auto endIt = d->m_recentActions.begin() + difference;
+        std::for_each(beginIt, endIt, [this](const Info &info) {
+            // Remove the action from the menus, action groups ...etc
+            delete removeAction(info.action);
+        });
+
+        // Erase the respective items from m_recentActions
+        d->m_recentActions.erase(beginIt, endIt);
     }
 }
 
@@ -123,7 +143,8 @@ static QString titleWithSensibleWidth(const QString &nameValue, const QString &v
         // If it does not fit, try to cut only the whole path, though if the
         // name is too long (more than 3/4 of the whole text) we cut it a bit too
         const int nameValueMaxWidth = maxWidthForTitles * 3 / 4;
-        QString cutNameValue, cutValue;
+        QString cutNameValue;
+        QString cutValue;
         if (nameWidth > nameValueMaxWidth) {
             cutNameValue = fontMetrics.elidedText(nameValue, Qt::ElideMiddle, nameValueMaxWidth);
             cutValue = fontMetrics.elidedText(value, Qt::ElideMiddle, maxWidthForTitles - nameValueMaxWidth);
@@ -136,7 +157,7 @@ static QString titleWithSensibleWidth(const QString &nameValue, const QString &v
     return title;
 }
 
-void KRecentFilesAction::addUrl(const QUrl &_url, const QString &name)
+void KRecentFilesAction::addUrl(const QUrl &url, const QString &name)
 {
     Q_D(KRecentFilesAction);
 
@@ -145,47 +166,41 @@ void KRecentFilesAction::addUrl(const QUrl &_url, const QString &name)
         return;
     }
 
-    /**
-     * Create a deep copy here, because if _url is the parameter from
-     * urlSelected() signal, we will delete it in the removeAction() call below.
-     * but access it again in the addAction call... => crash
-     */
-    const QUrl url(_url);
-
     if (url.isLocalFile() && url.toLocalFile().startsWith(QDir::tempPath())) {
         return;
     }
-    const QString tmpName = name.isEmpty() ? url.fileName() : name;
-    const QString pathOrUrl(url.toDisplayString(QUrl::PreferLocalFile));
 
+    using Info = KRecentFilesActionPrivate::RecentActionInfo;
+    // Reverse iterators because the most recent actions are at the end of m_recentActions
+    auto it = std::find_if(d->m_recentActions.rbegin(), d->m_recentActions.rend(), [&url](const Info &info) {
+        return info.url == url;
+    });
+    if (it != d->m_recentActions.rend()) { // url already has an action in the menu
+        // Remove the action...
+        QAction *act = removeAction(it->action);
+        // ... move the relevant object to the end of m_recentActions
+        std::rotate(d->m_recentActions.rbegin(), it, it + 1);
+        // ... prepend the action to the menu
+        menu()->insertAction(menu()->actions().value(0), act);
+        return; // All done
+    }
+
+    // Remove oldest item if already maxItems in list
+    Q_ASSERT(d->m_maxItems > 0);
+    if (static_cast<int>(d->m_recentActions.size()) == d->m_maxItems) {
+        auto oldest = d->m_recentActions.begin();
+        delete removeAction(oldest->action);
+        d->m_recentActions.erase(oldest);
+    }
+
+    const QString pathOrUrl(url.toDisplayString(QUrl::PreferLocalFile));
+    const QString tmpName = !name.isEmpty() ? name : url.fileName();
 #ifdef Q_OS_WIN
     const QString file = url.isLocalFile() ? QDir::toNativeSeparators(pathOrUrl) : pathOrUrl;
 #else
     const QString file = pathOrUrl;
 #endif
 
-    // remove file if already in list
-    const auto lstActions = selectableActionGroup()->actions();
-    for (QAction *action : lstActions) {
-        const QString urlStr = d->m_urls[action].toDisplayString(QUrl::PreferLocalFile);
-#ifdef Q_OS_WIN
-        const QString tmpFileName = url.isLocalFile() ? QDir::toNativeSeparators(urlStr) : urlStr;
-        if (tmpFileName.endsWith(file, Qt::CaseInsensitive))
-#else
-        if (urlStr.endsWith(file))
-#endif
-        {
-            removeAction(action)->deleteLater();
-            break;
-        }
-    }
-    // remove oldest item if already maxitems in list
-    Q_ASSERT(d->m_maxItems > 0);
-    if (selectableActionGroup()->actions().count() == d->m_maxItems) {
-        // remove oldest added item
-        delete removeAction(selectableActionGroup()->actions().first());
-    }
-
     d->m_noEntriesAction->setVisible(false);
     d->clearSeparator->setVisible(true);
     d->clearAction->setVisible(true);
@@ -201,35 +216,40 @@ void KRecentFilesAction::addAction(QAction *action, const QUrl &url, const QStri
     Q_D(KRecentFilesAction);
 
     menu()->insertAction(menu()->actions().value(0), action);
-    d->m_shortNames.insert(action, name);
-    d->m_urls.insert(action, url);
+    d->m_recentActions.push_back({action, url, name});
 }
 
 QAction *KRecentFilesAction::removeAction(QAction *action)
 {
-    Q_D(KRecentFilesAction);
-    KSelectAction::removeAction(action);
-
-    d->m_shortNames.remove(action);
-    d->m_urls.remove(action);
-
-    return action;
+    return KSelectAction::removeAction(action);
 }
 
 void KRecentFilesAction::removeUrl(const QUrl &url)
 {
     Q_D(KRecentFilesAction);
-    for (QMap<QAction *, QUrl>::ConstIterator it = d->m_urls.constBegin(); it != d->m_urls.constEnd(); ++it)
-        if (it.value() == url) {
-            delete removeAction(it.key());
-            return;
-        }
+
+    auto it = d->findByUrl(url);
+
+    if (it != d->m_recentActions.cend()) {
+        delete removeAction(it->action);
+        d->m_recentActions.erase(it);
+    };
 }
 
 QList<QUrl> KRecentFilesAction::urls() const
 {
     Q_D(const KRecentFilesAction);
-    return d->m_urls.values();
+
+    QList<QUrl> list;
+    list.reserve(d->m_recentActions.size());
+
+    using Info = KRecentFilesActionPrivate::RecentActionInfo;
+    // Reverse order to match how the actions appear in the menu
+    std::transform(d->m_recentActions.crbegin(), d->m_recentActions.crend(), std::back_inserter(list), [](const Info &info) {
+        return info.url;
+    });
+
+    return list;
 }
 
 void KRecentFilesAction::clear()
@@ -242,8 +262,7 @@ void KRecentFilesAction::clearEntries()
 {
     Q_D(KRecentFilesAction);
     KSelectAction::clear();
-    d->m_shortNames.clear();
-    d->m_urls.clear();
+    d->m_recentActions.clear();
     d->m_noEntriesAction->setVisible(true);
     d->clearSeparator->setVisible(false);
     d->clearAction->setVisible(false);
@@ -263,10 +282,13 @@ void KRecentFilesAction::loadEntries(const KConfigGroup &_config)
     QUrl url;
 
     KConfigGroup cg = _config;
-    if (cg.name().isEmpty()) {
+    // "<default>" means the group was constructed with an empty name
+    if (cg.name() == QLatin1String("<default>")) {
         cg = KConfigGroup(cg.config(), "RecentFiles");
     }
 
+    std::set<QUrl> seenUrls;
+
     bool thereAreEntries = false;
     // read file list
     for (int i = 1; i <= d->m_maxItems; i++) {
@@ -282,8 +304,9 @@ void KRecentFilesAction::loadEntries(const KConfigGroup &_config)
             continue;
         }
 
-        // Don't restore where the url is already known (eg. broken config)
-        if (d->m_urls.values().contains(url)) {
+        auto [it, isNewUrl] = seenUrls.insert(url);
+        // Don't restore if this url has already been restored (e.g. broken config)
+        if (!isNewUrl) {
             continue;
         }
 
@@ -313,26 +336,22 @@ void KRecentFilesAction::loadEntries(const KConfigGroup &_config)
 void KRecentFilesAction::saveEntries(const KConfigGroup &_cg)
 {
     Q_D(KRecentFilesAction);
-    QString key;
-    QString value;
-    QStringList lst = items();
 
     KConfigGroup cg = _cg;
-    if (cg.name().isEmpty()) {
+    // "<default>" means the group was constructed with an empty name
+    if (cg.name() == QLatin1String("<default>")) {
         cg = KConfigGroup(cg.config(), "RecentFiles");
     }
 
     cg.deleteGroup();
 
     // write file list
-    for (int i = 1; i <= selectableActionGroup()->actions().count(); i++) {
-        key = QStringLiteral("File%1").arg(i);
-        // i - 1 because we started from 1
-        value = d->m_urls[selectableActionGroup()->actions()[i - 1]].toDisplayString(QUrl::PreferLocalFile);
-        cg.writePathEntry(key, value);
-        key = QStringLiteral("Name%1").arg(i);
-        value = d->m_shortNames[selectableActionGroup()->actions()[i - 1]];
-        cg.writePathEntry(key, value);
+    int i = 1;
+    for (const auto &[action, url, shortName] : d->m_recentActions) {
+        cg.writePathEntry(QStringLiteral("File%1").arg(i), url.toDisplayString(QUrl::PreferLocalFile));
+        cg.writePathEntry(QStringLiteral("Name%1").arg(i), shortName);
+
+        ++i;
     }
 }
 
